@@ -5,6 +5,7 @@
 #include "base.h"
 #include "array.h"
 #include "buffer.h"
+#include "burl.h"       /* HTTP_PARSEOPT_HEADER_STRICT */
 #include "chunk.h"
 #include "fdevent.h"
 #include "log.h"
@@ -29,6 +30,12 @@
 #include "sys-socket.h"
 #include "sys-unistd.h" /* <unistd.h> */
 
+#ifndef _WIN32
+#include <arpa/inet.h>  /* htonl() */
+#else
+#include <winsock2.h>   /* htonl() */
+#endif
+
 /**
  * max size of the HTTP response header from backends
  * (differs from server.max-request-field-size for max request field size)
@@ -403,7 +410,7 @@ void http_response_send_file (request_st * const r, const buffer * const path, s
 		if (sce->fd < 0 && __builtin_expect( (0 != sce->st.st_size), 0)) {
 			r->http_status = (errno == ENOENT) ? 404 : 403;
 			if (r->conf.log_request_handling) {
-				log_perror(r->conf.errh, __FILE__, __LINE__,
+				log_pdebug(r->conf.errh, __FILE__, __LINE__,
 				  "file open failed: %s", path->ptr);
 			}
 			return;
@@ -414,9 +421,9 @@ void http_response_send_file (request_st * const r, const buffer * const path, s
 	    && 0 != stat_cache_path_contains_symlink(path, r->conf.errh)) {
 		r->http_status = 403;
 		if (r->conf.log_request_handling) {
-			log_error(r->conf.errh, __FILE__, __LINE__,
+			log_debug(r->conf.errh, __FILE__, __LINE__,
 			  "-- access denied due symlink restriction");
-			log_error(r->conf.errh, __FILE__, __LINE__,
+			log_debug(r->conf.errh, __FILE__, __LINE__,
 			  "Path         : %s", path->ptr);
 		}
 		return;
@@ -426,7 +433,7 @@ void http_response_send_file (request_st * const r, const buffer * const path, s
 	if (__builtin_expect( (!S_ISREG(sce->st.st_mode)), 0)) {
 		r->http_status = 403;
 		if (r->conf.log_file_not_found) {
-			log_error(r->conf.errh, __FILE__, __LINE__,
+			log_debug(r->conf.errh, __FILE__, __LINE__,
 			  "not a regular file: %s -> %s",
 			  r->uri.path.ptr, path->ptr);
 		}
@@ -466,11 +473,11 @@ void http_response_send_file (request_st * const r, const buffer * const path, s
 
 		const buffer * const lmod =
 		  __builtin_expect( (!light_btst(r->resp_htags, HTTP_HEADER_LAST_MODIFIED)), 1)
-		  ? http_response_set_last_modified(r, sce->st.st_mtime)
+		  ? http_response_set_last_modified(r, TIME64_CAST(sce->st.st_mtime))
 		  : NULL;
 
 		if (http_response_maybe_cachable(r)
-		    && HANDLER_FINISHED == http_response_handle_cachable(r, lmod, sce->st.st_mtime))
+		    && HANDLER_FINISHED == http_response_handle_cachable(r, lmod, TIME64_CAST(sce->st.st_mtime)))
 			return;
 	}
 
@@ -858,7 +865,7 @@ static int http_response_append_splice(request_st * const r, http_response_opts
 #endif
 
 
-static int http_response_append_mem(request_st * const r, const char * const mem, size_t len) {
+static int http_response_append_mem(request_st * const r, const char * const mem, uint32_t len) {
     if (r->resp_decode_chunked)
         return http_chunk_decode_append_mem(r, mem, len);
 
@@ -952,12 +959,29 @@ static int http_response_process_headers(request_st * const restrict r, http_res
         }
     }
     else if (__builtin_expect( (opts->backend == BACKEND_PROXY), 0)) {
-        /* invalid response Status-Line from HTTP proxy */
+        /* invalid response Status-Line from HTTP backend */
         r->http_status = 502; /* Bad Gateway */
         r->handler_module = NULL;
         return 0;
     }
 
+    const unsigned int http_header_strict =
+      (r->conf.http_parseopts & HTTP_PARSEOPT_HEADER_STRICT);
+
+    /* strictly validate all lines end "\r\n" for BACKEND_PROXY */
+    if (opts->backend == BACKEND_PROXY && http_header_strict) {
+        for (int j = 1; j <= hoff[0]; ++j) {
+            /*(nph check above ensures at least 2 chars on first line)*/
+            if (s[hoff[i+1]-2] != '\r') {
+                log_error(r->conf.errh, __FILE__, __LINE__,
+                  "HTTP backend response missing CR before LF");
+                r->http_status = 502; /* Bad Gateway */
+                r->handler_module = NULL;
+                return 0;
+            }
+        }
+    }
+
     for (; i < hoff[0]; ++i) {
         const char *k = s+hoff[i], *value;
         char *end = s+hoff[i+1]-1; /*('\n')*/
@@ -995,11 +1019,16 @@ static int http_response_process_headers(request_st * const restrict r, http_res
             }
             else if (id == HTTP_HEADER_OTHER && klen > 9 && (k[0] & 0xdf) == 'V'
                      && buffer_eq_icase_ssn(k, CONST_STR_LEN("Variable-"))) {
+                /* trust response from configured authorizer; skip validating */
                 http_header_env_append(r, k + 9, klen - 9, value, end - value);
             }
             continue;
         }
 
+        if (NULL != http_request_field_check_value(value, end - value,
+                                                   http_header_strict))
+            continue; /* invalid char in field value; skip */
+
         switch (id) {
           case HTTP_HEADER_STATUS:
             if (opts->backend != BACKEND_PROXY) {
@@ -1009,7 +1038,8 @@ static int http_response_process_headers(request_st * const restrict r, http_res
                   #ifdef __COVERITY__ /* Coverity false positive for tainted */
                     status = 200;/* http_header_str_to_code() validates */
                   #endif
-                    r->http_status = status;
+                    if (0 == r->http_status || 200 == r->http_status)
+                        r->http_status = status;
                     opts->local_redir = 0; /*(disable; status was set)*/
                 }
                 else {
@@ -1021,10 +1051,10 @@ static int http_response_process_headers(request_st * const restrict r, http_res
             break;
           case HTTP_HEADER_UPGRADE:
             /*(technically, should also verify Connection: upgrade)*/
-            /*(flag only for mod_proxy and mod_cgi (for now))*/
-            if (opts->backend != BACKEND_PROXY && opts->backend != BACKEND_CGI)
-                continue;
-            if (r->http_version >= HTTP_VERSION_2 && !r->h2_connect_ext)
+            /* likely broken behavior if backend thinks connection is being
+             * upgraded but client does not receive Upgrade, but if policy
+             * disallowed Upgrade, strip out Upgrade response header */
+            if (!opts->upgrade || r->http_status != 101)
                 continue;
             break;
           case HTTP_HEADER_CONNECTION:
@@ -1070,6 +1100,14 @@ static int http_response_process_headers(request_st * const restrict r, http_res
             }
             break;
           case HTTP_HEADER_TRANSFER_ENCODING:
+            /* strictly accept only Transfer-Encoding: chunked
+             * Technically, could allow "identity, chunked" and related,
+             * but still should reject unrecognized encodings */
+            if (!buffer_eq_icase_ss(value,end-value,CONST_STR_LEN("chunked"))) {
+                r->http_status = 502; /* Bad Gateway */
+                r->handler_module = NULL;
+                continue;
+            }
             if (light_btst(r->resp_htags, HTTP_HEADER_CONTENT_LENGTH)) {
                 /* ignore Content-Length if Transfer-Encoding: chunked
                  * (might choose to treat this as 502 Bad Gateway) */
@@ -1077,7 +1115,6 @@ static int http_response_process_headers(request_st * const restrict r, http_res
                 http_header_response_unset(r, HTTP_HEADER_CONTENT_LENGTH,
                                            CONST_STR_LEN("Content-Length"));
             }
-            /*(assumes "Transfer-Encoding: chunked"; does not verify)*/
             r->resp_decode_chunked = 1;
             r->gw_dechunk = ck_calloc(1, sizeof(response_dechunk));
             continue;
@@ -1087,11 +1124,9 @@ static int http_response_process_headers(request_st * const restrict r, http_res
             /* (not bothering to remove HTTP2-Settings from Connection) */
             continue;
           case HTTP_HEADER_OTHER:
-            /* ignore invalid headers with whitespace between label and ':'
-             * (if less strict behavior is desired, check and correct above
-             *  this switch() statement, but not for BACKEND_PROXY) */
-            if (k[klen-1] == ' ' || k[klen-1] == '\t')
-                continue;
+            if (NULL != http_request_field_check_name(k, klen,
+                                                      http_header_strict))
+                continue; /* invalid char in field name; skip */
             break;
           default:
             break;
@@ -1238,6 +1273,31 @@ handler_t http_response_parse_headers(request_st * const r, http_response_opts *
         return HANDLER_FINISHED;
     }
 
+    if (!opts->authorizer && light_btst(r->resp_htags, HTTP_HEADER_UPGRADE)) {
+        /*(already checked in http_response_process_headers())*/
+        /*if (opts->upgrade && r->http_status == 101)*/
+        {
+            /* 101 Switching Protocols; transition to transparent proxy */
+            opts->upgrade = 2;
+            if (r->h2_connect_ext) {
+                r->http_status = 200; /* OK (response status for CONNECT) */
+                http_header_response_unset(r, HTTP_HEADER_UPGRADE,
+                                           CONST_STR_LEN("Upgrade"));
+                http_header_response_unset(r, HTTP_HEADER_OTHER,
+                                         CONST_STR_LEN("Sec-WebSocket-Accept"));
+            }
+            http_response_upgrade_read_body_unknown(r);
+        }
+    }
+    else if (__builtin_expect( (r->h2_connect_ext != 0), 0)
+             && r->http_status < 300) {
+        /*(not handling other 1xx intermediate responses here; not expected)*/
+        http_response_body_clear(r, 0);
+        r->handler_module = NULL;
+        r->http_status = 405; /* Method Not Allowed */
+        return HANDLER_FINISHED;
+    }
+
     if (opts->local_redir && r->http_status >= 300 && r->http_status < 400
         && 0 == blen) {
         /* (Might not have begun to receive body yet, but do skip local-redir
